///////////////////////////////////////////////////////////////////
// J-ACE (Jaka Arcade Controls Encoder) v0.7 for HW v.1.0
// Copyright 2020 Jaakko Kairus
// jaakko@kair.us
//
// You may freely use and modify this software for personal use.
// Commercial use of this software or parts of it is prohibited
// without written permission from author.
// DISCLAIMER: This software comes with ABSOLUTELY NO WARRANTY
//
// For more information about this project see project web page at
// http://kair.us/projects/j_ace/
//
// Compiled with CCS compiler version 5.092
// Total compiled length 4388 words for keyboard mode
// Total compiled length 3895 words for joystick mode
// Program lengths include 1024 word bootloader
///////////////////////////////////////////////////////////////////
// Recent changes:
// v0.1 13.2.2020
// Initial version. Enumerates as 3 HID boot protocol keyboards
// Keys dedicated to a certain keyboard. Length 3486 words.
// v0.2 15.2.2020
// Now all keys are prioritised to first keyboard. 2nd and 3rd
// are used if all keys don't fit to first report. Length 3478 words.
// v0.3 19.2.2020
// Added switch debounce function. Length 3660 words.
// v0.4 19.2.2020
// Added selectable keymaps for players 1,2,3 or 4.
// Added 'shift' function for player 1 keymap, start button works as shift
// and changes keycodes of joystick and action buttons
// Length 3784 words
// v0.5 30.3.2020
// Changed debouncne function to use timer so that it
// is independent of USB polling rate. 3959 words
// 420 us for button readout and debounce
// 350 us for just the debounce
// Verified that actual polling interval is still 1ms even if all 3
// keyboards are sending data.
// Should add a way to either rescan buttons until USB polls data
// or wait for known time before reading buttons so that USB poll
// will happen soon after it, to minimize delay especially if poll
// rate is less than 1000 Hz
// Added autofire. 
// Autofire has configurable speed and buttons. Autofire LED
// indicates speed, activation and when in configure mode.
// Autofire speed adjusts with polling rare. Length 4109 words.
// 420us total for usb_keyboard_task() before sending data
// 500us when global autofire is on
// each autofire button down adds ca. 10 us.
// Added saving of autofire speed and autofire buttons.
// Added player/keymap setup mode (keep button 1 pressed when
// plugging in USB cable). Player/keymap also saved.
// Checking validity of settings saved to EEPROM. Length 4375
// words.
// v0.6 1.4.2020
// Splitted usb_keyboard_task() into several separate functions
// Allows to enable/disable different functions easily, makes
// code easier to read and prepares to add also joystick mode
// in addition to keyboard encoder mode. Length 4387 words.
// v0.7 5.4.2020
// Added joystick mode. Define J_ACE_JOYSTICK to compile
// joystick mode firmware. Joystick mode also implements the
// 'shift' function for 15 additional buttons. You can use
// the keymap setup mode to set player to other than 1 to
// disable the shift function.

//#define J_ACE_JOYSTICK

#include "j_ace.h"

void main(void) {
	usb_init();
	PORT_E_PULLUPS(TRUE);					// Activates weak pull-up on pin RA3
	PORT_B_PULLUPS(0b11111111);				// Activates weak pull-up on all port B pins
	set_tris_c(0x3c);						// Set bits 0,1,6,7 as LED outputs. bit 2 is AUTOFIRE input

	setup_timer_1(T1_INTERNAL|T1_DIV_BY_8);	// Setup timer1 to use fclk/4 (12 MHz) as input, prescaler 8. One tick is 666.667ns
											// Thus 1500 ticks / millisecond
	load_jace_settings();					// Load settings from EEPROM (Autofire mode, keymap etc.)
	setup_keymap();							// Run keymap setup mode if button 1 is pressed during USB plug-in

	// Main loop, run it forever
	while (TRUE) {
		if (usb_enumerated()) {
			measure_polling_interval();		// Measures USB polling interval and updates time constants for debounce and autofire
			read_debounced_inputs();		// Reads and debounces inputs. Debounced data stored in active_buttons[] array
			handle_shift();					// Handles 'shift' function, which means that alternate keymap is used when start button is held down
			memcpy(output_buttons,active_buttons,sizeof(active_buttons));	// Copy the active buttons to output array
			add_autofire();					// Adds autofire function. Includes also autofire setup mode.
			#ifdef	J_ACE_JOYSTICK
			usb_joystick_task();
			#else
			send_keycodes_trplkbd();		// Converts output_buttons[] array to keycodes and sends data via three HID keyboards
			#endif
		}
	}
}

int1 debounce_idle(unsigned int8 *debounce_timer) {
	for (int i=0;i<17;i++) {
		if (debounce_timer[i] != 0)
			return 0;
	}
	return 1;
}	

void add_autofire() {	
	////////////////////////////////////////
	/// AUTOFIRE functions
	////////////////////////////////////////
	static int1 global_autofire = 0;
	static int1 autofire_high = 0;
	static int1 autofire_setup_mode = 0;
	static unsigned int16 autofire_activation_counter = 0;
	static unsigned int8 autofire_delay_counter = 0;
	// moved to global static unsigned int8 autofire_speed_setting = 32;

	// Autofire button press length detection counter:
	if (active_buttons[16]) {
		autofire_activation_counter+=poll_length_ms;
	}
	else {
		autofire_activation_counter = 0;
	}
	
	// Normal autofire activation/deactivaion:
	if (autofire_activation_counter == poll_length_ms)
		global_autofire = !global_autofire;
	
	// Autofire setup mode activation
	if (autofire_activation_counter >= 1024) {
		autofire_setup_mode = 1;
		global_autofire = 0;
	}	
	
	// Autofire speed counter
	autofire_delay_counter += poll_length_ms;
	if (autofire_delay_counter >= autofire_speed_setting) {
		autofire_delay_counter = 0;
		autofire_high = !autofire_high;
	}	
	
	// Autofire indication LED
	if (global_autofire) {
		output_low(PIN_LED_4);
	}
	else {
		output_high(PIN_LED_4);					
	}
	
	// Autofire operation on activated buttons
	if (global_autofire) {
		for (unsigned int8 i = 0; i<16; i++) {
			if (active_buttons[i]) {
				if (autofire_buttons[i]) {
					if (previous_buttons[i]==0) {	// This ensures that button reacts immediately
						autofire_delay_counter=0;
						autofire_high=1;
					}	
					if (autofire_high) {
						output_buttons[i] = 1;
						output_low(PIN_LED_4);
					}
					else {
						output_buttons[i] = 0;
						output_high(PIN_LED_4);
					}			
				}
			}
		}
	}				

	// Autofire setup
	if (autofire_setup_mode) {
		if (autofire_high)				// indicates current autofire speed by autofire led
			output_low(PIN_LED_4);		// this also indicates that autofire setup mode is active
		else
			output_high(PIN_LED_4);
			
		if (autofire_activation_counter == poll_length_ms) {
			autofire_setup_mode = 0;
			global_autofire = 0;
			memcpy(autofire_buttons,active_buttons,sizeof(autofire_buttons));	// Copy the active buttons to autofire buttons array when exiting autofire setup
			write_eeprom(1,autofire_speed_setting);
			write_eeprom(2,*((unsigned int8 *)autofire_buttons));
			write_eeprom(3,*((unsigned int8 *)(autofire_buttons)+1));
		}
		else {
			if (active_buttons[8] && !previous_buttons[8]) {	// joystick up -> increase autofire speed
				if (autofire_speed_setting > AUTOFIRE_DELAY_LOW_LIMIT) {
					autofire_speed_setting /= 2;
				}
			}
			if (active_buttons[9] && !previous_buttons[9]) {	// joystick down -> decrease autofire speed
				if (autofire_speed_setting < AUTOFIRE_DELAY_HIGH_LIMIT) {
					autofire_speed_setting *= 2;
				}
			}
		}
	}
}

void send_keycodes_trplkbd() {
	unsigned int8 tx_msg[20];
	unsigned int8 leds;
	unsigned int8 butindex=2;

    memset(tx_msg, 0x00, sizeof(tx_msg));
			
	// update tx_msg with keyboard scancodes for currently pressed down buttons
	for (int i=0;i<16;i++) {
		if (output_buttons[i]) {
			if (kbd_scancodes[i][player+shift_active+1])					// The scancode for this button is a modifier key
				tx_msg[0] |= kbd_scancodes[i][player+shift_active];
			else											// The scancode for this button is a normal key
				tx_msg[butindex++] = kbd_scancodes[i][player+shift_active];
		}
	}
	
	// For testing delay:
	//if (active_buttons[16])
	//	tx_msg[butindex++] = KEY_NUMLOCK;

	while(!usb_put_packet(1, tx_msg, 8, USB_DTS_TOGGLE)){}	// if only one of the usb_put_packet() has while, it must be this first one!
															// otherwise this will be skipped and data is sent on next round

   //receive NUM LOCK, CAPS LOCK, etc LED status from PC.
   if (usb_kbhit(1)) {
      usb_get_packet(1, &leds, 1);
      leds = (leds & 0x03) | ((leds & 0x0c)<<4);
      output_c(leds ^ 0xff);		// invert bits, led must be connected between +5V and LED output pin
   }
   
   // Send the remaining buttons (if any)
    memset(tx_msg+6, 0x00, 2);
   	while(!usb_put_packet(2, tx_msg+6, 8, USB_DTS_TOGGLE)){}
    memset(tx_msg+12, 0x00, 2);
   	while(!usb_put_packet(3, tx_msg+12, 8, USB_DTS_TOGGLE)){}
}

void handle_shift() {
	// Handle 'shift' function if configured as player 1 keyboard
	static int1 shift_used = 0;
	static int1 shift_allowed = 0;
	static unsigned int8 shift_counter = 0;
	shift_active = 0;

	if (player == 0) {
		if ((current_startbutton == 0) && (previous_buttons[12] == 1) && (shift_used == 0))	{	// Activate start button when it is released
			shift_counter = 50;
		}
		if ((current_startbutton == 1) && (previous_buttons[12] == 0)) {		// When start button is pressed,
			shift_used = 0;														// start monitorin if it is used as shift
		}
		if ((current_startbutton == 1) && (previous_buttons[12] == 0)) {		// Shift is allowed only if no other keys are being
			if((*(unsigned int16*)active_buttons) & 0xefff)						// pressed when start button is pressed
				shift_allowed = 0;												// otherwise it wouldn't be possible to e.g.
			else																// continue game in 1942 because it requires
				shift_allowed = 1;												// fire being held down while pressing start
		}	
		if (current_startbutton == 1 && shift_allowed) {						// When start button is kept down, activate shift function
			active_buttons[12] = 0;												// and clear the original active_buttons start button bit
			shift_active = 8;
		}
		if ((current_startbutton == 1) && *(unsigned int16*)active_buttons && shift_allowed) {	// If start button is used as shift, do not send
			shift_used = 1;														// normal start button scancode!
		}
		if (shift_counter) {													// Keep start buttons pressed some time after release
			active_buttons[12] = 1;
			shift_counter--;
		}
	}
}

void read_debounced_inputs() {
	unsigned int32 input_shadow;

	input_shadow = make32(input(PIN_AUTOFIRE),input_b(),input_a());
	
	active_buttons[12] = current_startbutton;
	memcpy(previous_buttons,active_buttons,sizeof(active_buttons));
	
	// Debounce all port A, port B inputs and autofire input
	for (int i=0;i<17;i++) {
		if ((!bit_test(input_shadow,i) || debounce_timer[i]!=0) && debounce_timer[i] < press_debounce_polls)
			debounce_timer[i]++;
		
		if ((bit_test(input_shadow, i) && debounce_timer[i] == press_debounce_polls) || debounce_timer[i] > press_debounce_polls)
			debounce_timer[i]++;
			
		if (debounce_timer[i] == total_debounce_polls)
			debounce_timer[i] = 0;
		
		if (debounce_timer[i] == 1)
			active_buttons[i] = 1;
		
		if (debounce_timer[i] == press_debounce_polls + 1)
			active_buttons[i] = 0;
	}
	current_startbutton = active_buttons[12];
}

void measure_polling_interval() {
	static unsigned int16 poll_interval_timer = 60000;	// do the first poll_length measurement ca. 5 seconds after enumeration
	static unsigned int16 poll_length = 0;
	static int1 new_poll_length = 0;

	// Measure host polling interval
	poll_interval_timer++;
	if (poll_interval_timer<10) {
		poll_length += (get_timer1() / 10);	// poll_length should be able to handle total of 430 ms
		set_timer1(0);
	}
	else if (poll_interval_timer==10) {
		poll_length = (poll_length + 75) / 150;	// poll_length is now milliseconds for 10 rounds
		new_poll_length = 1;
	}
	else if (poll_interval_timer==65535) {
		new_poll_length = 0;
		poll_length = 0;
		set_timer1(0);
	}	
	
	// Update debounce timer time constants, but only if debounce algorithm is idle (no button presses being processed)
	if (new_poll_length) {
		if (debounce_idle(debounce_timer)) {
			press_debounce_polls = (unsigned int8)(PRESS_DEBOUNCE_MS * 10 / poll_length);
			total_debounce_polls = (unsigned int8)(RELEASE_DEBOUNCE_MS * 10 / poll_length + press_debounce_polls);
			// Handle cases where polling interval w.r.t. requested debounce time is so long that it messes the debounce function
			if (press_debounce_polls<2)
				press_debounce_polls=2;
			if (total_debounce_polls<press_debounce_polls+2)
				total_debounce_polls=press_debounce_polls+2;
			new_poll_length = 0;
			poll_length_ms = poll_length / 10;
		}
	}		
}

void load_jace_settings() {
	// Load J-ACE settings from EEPROM
	player = read_eeprom(0);
	autofire_speed_setting = read_eeprom(1);
	*((unsigned int8 *)autofire_buttons) = read_eeprom(2);
	*((unsigned int8 *)(autofire_buttons)+1) = read_eeprom(3);
	
	// Check validity of EEPROM data
	if (player>6)
		player=0;
	if (autofire_speed_setting > AUTOFIRE_DELAY_HIGH_LIMIT)
		autofire_speed_setting = AUTOFIRE_DEFAULT_DELAY;
	if (autofire_speed_setting < AUTOFIRE_DELAY_LOW_LIMIT)
		autofire_speed_setting = AUTOFIRE_DEFAULT_DELAY;
	if (autofire_buttons[8] && autofire_buttons[9]) // Joystick up and down
		*((unsigned int8 *)(autofire_buttons)+1) = 0x00;	
}

void setup_keymap() {
		// Setup player / keymap
	int i=0;
	if (!input(PIN_BUT_1)) {				// Player (keymap) selection entered by keeping button 1 pressed when plugging in USB
		while(!input(PIN_BUT_1)) {}			// wait release of button 1
		delay_ms(RELEASE_DEBOUNCE_MS);		// deboune the release
		while(input(PIN_BUT_1)) {			// wait again press of button 1
			i++;
			delay_ms(1);
			if (i<127)						// indicate setup mode by flashing leds
				output_c(0x00);
			else
				output_c(0xff);
		}
		if (!input(PIN_JOY_UP))
			player = 0;
		if (!input(PIN_JOY_RIGHT))
			player = 2;
		if (!input(PIN_JOY_DOWN))
			player = 4;
		if (!input(PIN_JOY_LEFT))
			player = 6;
		write_eeprom(0, player);
	}				
}

void usb_joystick_task(void) 
{
	int8 out_data[6];

    memset(out_data, 0x00, sizeof(out_data));

	if (shift_active) {
		out_data[1] = *((unsigned int8 *)output_buttons+1) << 4;					// Joystick directions as buttons 13..16
		out_data[2] = *((unsigned int8 *)output_buttons);							// Buttons 1..8 as 17..24		
		out_data[2] = out_data[2]&0x3f | out_data[2]>>1&0x40 | out_data[2]<<1&0x80;	// Swap buttons 7 and 8
		out_data[3] = *((unsigned int8 *)output_buttons+1) >> 5 & 0x07;				// coin, a and b as buttons 25..27
	}
	else {
		out_data[0] = *((unsigned int8 *)output_buttons);
		out_data[0] = out_data[0]&0x3f | out_data[0]>>1&0x40 | out_data[0]<<1&0x80;	// Swap buttons 7 and 8
		out_data[1] = *((unsigned int8 *)output_buttons+1) >> 4;					// Start, coin, A and B

		if(output_buttons[IDX_JOY_UP])				// up
			out_data[5]=-126;
		else if(output_buttons[IDX_JOY_DOWN])		// down
			out_data[5]=126;
	
	    if(output_buttons[IDX_JOY_LEFT])			// left
			out_data[4]=-126;
		else if(output_buttons[IDX_JOY_RIGHT])		// right
			out_data[4]=126;
	}	

	while (!usb_put_packet(1,out_data,6,USB_DTS_TOGGLE))	// wait here for USB poll from host
	{
	}
}
